import uuid

from typing import List

import brewman.schemas.user as schemas

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy import delete, select, update
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.attendance import Attendance
from ..models.login_history import LoginHistory
from ..models.role import Role
from ..models.user import User
from ..models.user_role import user_role
from ..models.voucher import Voucher
from ..schemas.user import UserToken


router = APIRouter()


@router.post("", response_model=schemas.User)
def save(
    data: schemas.UserIn,
    user: UserToken = Security(get_user, scopes=["users"]),
) -> schemas.User:
    try:
        with SessionFuture() as db:
            item = User(name=data.name, password=data.password, locked_out=data.locked_out)
            db.add(item)
            add_roles(item, data.roles, db)
            db.commit()
            return user_info(item, db, user)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("/me", response_model=schemas.User)
def show_me(
    user: UserToken = Depends(get_user),
) -> schemas.User:
    with SessionFuture() as db:
        item = db.execute(select(User).where(User.id == user.id_)).scalar_one()
        return user_info(item, db, user)


@router.put("/me", response_model=schemas.User)
def update_me(
    data: schemas.UserIn,
    user: UserToken = Depends(get_user),
) -> schemas.User:
    try:
        with SessionFuture() as db:
            item: User = db.execute(select(User).where(User.id == user.id_)).scalar_one()
            if "users" in user.permissions:
                item.name = data.name
                item.locked_out = data.locked_out
                add_roles(item, data.roles, db)
            if data.password and item.password != data.password:
                item.password = data.password
            db.commit()
            return user_info(item, db, user)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.put("/{id_}", response_model=schemas.User)
def update_route(
    id_: uuid.UUID,
    data: schemas.UserIn,
    user: UserToken = Security(get_user, scopes=["users"]),
) -> schemas.User:
    try:
        with SessionFuture() as db:
            item: User = db.execute(select(User).where(User.id == id_)).scalar_one()
            item.name = data.name
            if data.password and item.password != data.password:
                item.password = data.password
            item.locked_out = data.locked_out
            add_roles(item, data.roles, db)
            db.commit()
            return user_info(item, db, user)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def add_roles(user: User, roles: List[schemas.RoleItem], db: Session) -> None:
    for role in roles:
        ug = next((g for g in user.roles if g.id == role.id_), None)
        if role.enabled and ug is None:
            user.roles.append(db.execute(select(Role).where(Role.id == role.id_)).scalar_one())
        elif not role.enabled and ug:
            user.roles.remove(ug)


@router.delete("/{id_}", response_model=None)
def delete_route(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["users"]),
) -> None:
    if id_ == user.id_:
        raise HTTPException(
            status_code=status.HTTP_501_NOT_IMPLEMENTED,
            detail="You cannot delete yourself",
        )
    with SessionFuture() as db:
        item: User = db.execute(select(User).where(User.id == id_)).scalar_one()
        admin: User = db.execute(select(User).where(User.name.ilike("admin"))).scalar_one()
        db.execute(update(Voucher).where(Voucher.user_id == item.id).values(user_id=admin.id))
        db.execute(update(Attendance).where(Attendance.user_id == item.id).values(user_id=admin.id))
        db.execute(delete(LoginHistory).where(LoginHistory.user_id == item.id))
        db.execute(delete(user_role).where(user_role.c.user_id == item.id))
        db.execute(delete(User).where(User.id == item.id))
        db.commit()


@router.get("", response_model=schemas.UserBlank)
def show_blank(
    user: UserToken = Security(get_user, scopes=["users"]),
) -> schemas.UserBlank:
    with SessionFuture() as db:
        return user_blank(db)


@router.get("/list", response_model=List[schemas.UserList])
async def show_list(
    user: UserToken = Security(get_user, scopes=["users"]),
) -> List[schemas.UserList]:
    with SessionFuture() as db:
        users: List[User] = db.execute(select(User).order_by(User.name)).scalars().all()
        return [
            schemas.UserList(
                id=item.id,
                name=item.name,
                lockedOut=item.locked_out,
                roles=[p.name for p in sorted(item.roles, key=lambda p: p.name)],
                lastDevice=item.login_history[0].client.name if len(item.login_history) else "None",
                lastDate=item.login_history[0].date if len(item.login_history) else None,
            )
            for item in users
        ]


@router.get("/active", response_model=List[schemas.UserBlank])
async def show_active(user: UserToken = Depends(get_user)) -> List[schemas.UserBlank]:
    with SessionFuture() as db:
        return [
            schemas.UserBlank(name=name)
            for name in db.execute(select(User.name).where(User.locked_out == False).order_by(User.name))  # noqa: E712
            .scalars()
            .all()
        ]


@router.get("/{id_}", response_model=schemas.User)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["users"]),
) -> schemas.User:
    with SessionFuture() as db:
        item = db.execute(select(User).where(User.id == id_)).scalar_one()
        return user_info(item, db, user)


def user_info(item: User, db: Session, user: UserToken) -> schemas.User:
    return schemas.User(
        id=item.id,
        name=item.name,
        password="",
        lockedOut=item.locked_out,
        roles=[
            schemas.RoleItem(
                id=r.id,
                name=r.name,
                enabled=True if r in item.roles else False,
            )
            for r in db.execute(select(Role).order_by(Role.name)).scalars().all()
        ]
        if "advanced-delete" in user.permissions
        else [],
    )


def user_blank(db: Session) -> schemas.UserBlank:
    return schemas.UserBlank(
        name="",
        password="",
        lockedOut=False,
        roles=[
            schemas.RoleItem(id=r.id, name=r.name, enabled=False)
            for r in db.execute(select(Role).order_by(Role.name)).scalars().all()
        ],
    )
