import { Component, OnInit } from '@angular/core';
import { FormArray, FormControl, FormGroup } from '@angular/forms';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import * as moment from 'moment';
import { BehaviorSubject } from 'rxjs';
import { map } from 'rxjs/operators';

import { AuthService } from '../auth/auth.service';
import { Account } from '../core/account';
import { AccountBalance } from '../core/account-balance';
import { Incentive } from '../core/incentive';
import { ToasterService } from '../core/toaster.service';
import { User } from '../core/user';
import { Voucher } from '../core/voucher';
import { VoucherService } from '../core/voucher.service';
import { ConfirmDialogComponent } from '../shared/confirm-dialog/confirm-dialog.component';

import { IncentiveDataSource } from './incentive-datasource';

@Component({
  selector: 'app-incentive',
  templateUrl: './incentive.component.html',
  styleUrls: ['./incentive.component.css'],
})
export class IncentiveComponent implements OnInit {
  public incentiveObservable = new BehaviorSubject<Incentive[]>([]);
  dataSource: IncentiveDataSource = new IncentiveDataSource(this.incentiveObservable);
  form: FormGroup<{
    date: FormControl<Date>;
    incentives: FormArray<
      FormGroup<{
        points: FormControl<number>;
      }>
    >;
  }>;

  voucher: Voucher = new Voucher();
  account: Account = new Account();
  accBal: AccountBalance | null = null;

  displayedColumns = ['name', 'designation', 'department', 'daysWorked', 'points', 'amount'];

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dialog: MatDialog,
    private toaster: ToasterService,
    public auth: AuthService,
    private ser: VoucherService,
  ) {
    this.form = new FormGroup({
      date: new FormControl(new Date(), { nonNullable: true }),
      incentives: new FormArray<FormGroup<{ points: FormControl<number> }>>([]),
    });
    // Listen to Date Change
    this.form.controls.date.valueChanges
      .pipe(map((x) => moment(x).format('DD-MMM-YYYY')))
      .subscribe((x) => {
        if (x !== this.voucher.date && !this.voucher.id) {
          return this.ser.getIncentive(x).subscribe((voucher: Voucher) => {
            this.loadVoucher(voucher);
          });
        }
        return '';
      });
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as { voucher: Voucher };

      this.loadVoucher(data.voucher);
    });
  }

  loadVoucher(voucher: Voucher) {
    this.voucher = voucher;
    this.form.controls.date.setValue(moment(this.voucher.date, 'DD-MMM-YYYY').toDate());
    this.voucher.incentives.forEach((x) =>
      this.form.controls.incentives.push(
        new FormGroup({
          points: new FormControl(x.points, { nonNullable: true }),
        }),
      ),
    );
    this.dataSource = new IncentiveDataSource(this.incentiveObservable);
    this.incentiveObservable.next(this.voucher.incentives);
  }

  totalPoints() {
    return this.voucher.incentives
      .map((item) => item.daysWorked * item.points)
      .reduce((sum, item) => sum + item);
  }

  pointValue() {
    return Math.round(((this.voucher.incentive as number) * 100) / this.totalPoints()) / 100;
  }

  less(row: Incentive, i: number) {
    if (row.points >= 1) {
      row.points -= 1;
      this.form.controls.incentives.controls[i].setValue({
        points: row.points,
      });
    }
  }

  change(row: Incentive, i: number) {
    row.points = this.form.controls.incentives.controls[i].value.points ?? 0;
    this.form.controls.incentives.controls[i].setValue({ points: row.points });
  }

  more(row: Incentive, i: number) {
    row.points += 1;
    this.form.controls.incentives.controls[i].setValue({ points: row.points });
  }

  amount(row: Incentive) {
    return row.points * row.daysWorked * this.pointValue();
  }

  canSave() {
    if (!this.voucher.id) {
      return true;
    }
    if (this.voucher.posted && this.auth.allowed('edit-posted-vouchers')) {
      return true;
    }
    return (
      this.voucher.user.id === (this.auth.user as User).id ||
      this.auth.allowed("edit-other-user's-vouchers")
    );
  }

  post() {
    this.ser.post(this.voucher.id as string).subscribe(
      (result) => {
        this.loadVoucher(result);
        this.toaster.show('Success', 'Voucher Posted');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  save() {
    const voucher: Voucher = this.getVoucher();
    this.ser.saveOrUpdate(voucher).subscribe(
      (result) => {
        if (voucher.id === result.id) {
          this.loadVoucher(result);
        } else {
          this.router.navigate(['/incentive', result.id]);
        }
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  getVoucher(): Voucher {
    const formModel = this.form.value;
    this.voucher.date = moment(formModel.date).format('DD-MMM-YYYY');
    const array = this.form.controls.incentives;

    this.voucher.incentives.forEach((item, index) => {
      item.points = array.controls[index].value.points ?? 0;
    });
    return this.voucher;
  }

  delete() {
    this.ser.delete(this.voucher.id as string).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigate(['/incentive'], { replaceUrl: true });
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: { title: 'Delete Voucher?', content: 'Are you sure? This cannot be undone.' },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }
}
