import { AfterViewInit, Component, ElementRef, OnInit, ViewChild } from '@angular/core';
import { FormControl, FormGroup } from '@angular/forms';
import { MatAutocompleteSelectedEvent } from '@angular/material/autocomplete';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { round } from 'mathjs';
import * as moment from 'moment';
import { BehaviorSubject, Observable, of as observableOf } from 'rxjs';
import { debounceTime, distinctUntilChanged, map, switchMap } from 'rxjs/operators';

import { Product } from '../../core/product';
import { ProductSku } from '../../core/product-sku';
import { ToasterService } from '../../core/toaster.service';
import { ProductService } from '../../product/product.service';
import { ConfirmDialogComponent } from '../../shared/confirm-dialog/confirm-dialog.component';
import { MathService } from '../../shared/math.service';
import { Recipe } from '../recipe';
import { RecipeItem } from '../recipe-item';
import { RecipeService } from '../recipe.service';

import { RecipeDetailDatasource } from './recipe-detail-datasource';

@Component({
  selector: 'app-product-group-detail',
  templateUrl: './recipe-detail.component.html',
  styleUrls: ['./recipe-detail.component.css'],
})
export class RecipeDetailComponent implements OnInit, AfterViewInit {
  @ViewChild('productElement', { static: true }) productElement?: ElementRef;
  @ViewChild('ingredientElement', { static: true }) ingredientElement?: ElementRef;
  public itemsObservable = new BehaviorSubject<RecipeItem[]>([]);
  dataSource: RecipeDetailDatasource = new RecipeDetailDatasource(this.itemsObservable);
  form: FormGroup<{
    validFrom: FormControl<Date>;
    validTill: FormControl<Date>;
    recipeYield: FormControl<string | null>;
    costPrice: FormControl<string | null>;
    salePrice: FormControl<string | null>;
    costPercentage: FormControl<number | null>;
    product: FormControl<ProductSku | string | null>;
    addRow: FormGroup<{
      ingredient: FormControl<string | null>;
      quantity: FormControl<string>;
      rate: FormControl<string>;
    }>;
  }>;

  product: ProductSku | null;
  products: Observable<ProductSku[]>;
  ingredient: ProductSku | null;
  ingredients: Observable<ProductSku[]>;
  item: Recipe = new Recipe();

  displayedColumns = ['product', 'quantity', 'rate', 'amount', 'action'];

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dialog: MatDialog,
    private toaster: ToasterService,
    private math: MathService,
    private ser: RecipeService,
    private productSer: ProductService,
  ) {
    this.product = null;
    this.ingredient = null;
    this.form = new FormGroup({
      validFrom: new FormControl(new Date(), { nonNullable: true }),
      validTill: new FormControl(new Date(), { nonNullable: true }),
      recipeYield: new FormControl<string | null>(null),
      costPrice: new FormControl<string | null>(null),
      salePrice: new FormControl<string | null>(null),
      costPercentage: new FormControl<number | null>(null),
      product: new FormControl<ProductSku | string | null>(new ProductSku()),
      addRow: new FormGroup({
        ingredient: new FormControl(''),
        quantity: new FormControl('', { nonNullable: true }),
        rate: new FormControl('', { nonNullable: true }),
      }),
    });
    // Setup Product Autocomplete
    this.products = this.form.controls.product.valueChanges.pipe(
      map((x) => ((x as ProductSku).name !== undefined ? (x as ProductSku).name : (x as string))),
      map((x) => (x !== null && x.length >= 1 ? x : null)),
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) => (x === null ? observableOf([]) : this.productSer.autocompleteSku(x, false))),
    );
    // Setup Product Autocomplete
    this.ingredients = this.form.controls.addRow.controls.ingredient.valueChanges.pipe(
      map((x) => (x !== null && x.length >= 1 ? x : null)),
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) =>
        x === null ? observableOf([]) : this.productSer.autocompleteProduct(x, null),
      ),
    );
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as { item: Recipe };
      this.showItem(data.item);
      this.updateView();
    });
  }

  showItem(item: Recipe) {
    this.item = item;
    this.form.setValue({
      validFrom: moment(item.validFrom, 'DD-MMM-YYYY').toDate(),
      validTill: moment(item.validTill, 'DD-MMM-YYYY').toDate(),
      recipeYield: `${item.recipeYield}`,
      salePrice: `${item.salePrice}`,
      costPrice: `${item.costPrice}`,
      costPercentage: null,
      product: item.sku,
      addRow: {
        ingredient: null,
        quantity: '',
        rate: '',
      },
    });
    this.dataSource = new RecipeDetailDatasource(this.itemsObservable);
  }

  ngAfterViewInit() {
    setTimeout(() => {
      if (this.productElement) {
        this.productElement.nativeElement.focus();
      }
    }, 0);
  }

  displayIngredient(product?: Product): string {
    return product ? `${product.name} (${product.fractionUnits})` : '';
  }

  displayProduct(product?: Product): string {
    return product?.name || '';
  }

  productSelected(event: MatAutocompleteSelectedEvent): void {
    const product: ProductSku = event.option.value;
    this.item.sku = product;
    this.form.controls.product.setValue(product);
    this.form.controls.salePrice.setValue('' + (product.salePrice ?? 0));
  }

  ingredientSelected(event: MatAutocompleteSelectedEvent): void {
    const ingredient: ProductSku = event.option.value;
    this.ingredient = ingredient;
    const item = this.getItem();
    this.ser
      .getIngredientDetails(ingredient.id, item.validFrom, item.validTill)
      .subscribe((x) => this.form.controls.addRow.controls.rate.setValue('' + x.costPrice));
  }

  addRow() {
    const formValue = this.form.value.addRow;
    if (formValue === undefined) {
      return;
    }
    const quantity = this.math.parseAmount(formValue.quantity, 2);
    const rate = this.math.parseAmount(formValue.rate, 2);
    if (this.ingredient === null || quantity <= 0 || rate <= 0) {
      return;
    }
    const oldFiltered = this.item.items.filter(
      (x) => x.product.id === (this.ingredient as ProductSku).id,
    );
    if (oldFiltered.length) {
      this.toaster.show('Danger', 'Product already added');
      return;
    }
    this.item.items.push(
      new RecipeItem({
        product: this.ingredient,
        quantity,
        price: rate,
      }),
    );
    this.resetAddRow();
    this.updateView();
  }

  resetAddRow() {
    this.form.controls.addRow.reset();
    this.ingredient = null;
    setTimeout(() => {
      if (this.ingredientElement) {
        this.ingredientElement.nativeElement.focus();
      }
    }, 0);
  }

  updateView() {
    this.itemsObservable.next(this.item.items);
    const costPrice = round(
      this.item.items.map((x) => x.quantity * x.price).reduce((p, c) => p + c, 0),
      2,
    );
    this.form.controls.costPrice.setValue(`${costPrice}`);
    const salePrice = this.math.parseAmount(this.form.value.salePrice ?? '0', 2);
    if (salePrice < 0) {
      return;
    }
    const costPercentage = round((100 * costPrice) / salePrice, 2);
    this.form.controls.costPercentage.setValue(costPercentage);
  }

  deleteRow(row: RecipeItem) {
    this.item.items.splice(this.item.items.indexOf(row), 1);
    this.updateView();
  }

  save() {
    this.ser.saveOrUpdate(this.getItem()).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigateByUrl('/recipes');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  delete() {
    this.ser.delete(this.item.id as string).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigate(['/recipes']);
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: { title: 'Delete Recipe?', content: 'Are you sure? This cannot be undone.' },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  getItem(): Recipe {
    const formModel = this.form.value;
    this.item.validFrom = moment(formModel.validFrom).format('DD-MMM-YYYY');
    this.item.validTill = moment(formModel.validTill).format('DD-MMM-YYYY');
    this.item.recipeYield = this.math.parseAmount(formModel.recipeYield ?? '1', 2);
    this.item.salePrice = this.math.parseAmount(formModel.salePrice ?? '0', 2);
    this.item.costPrice = this.math.parseAmount(formModel.costPrice ?? '0', 2);
    return this.item;
  }
}
